//------------------------------------------------------------------------------
//
//  npgldraw.c
//
// ANTZ - 3D data viewer with realtime real-world interaction based on NPE.
//
// ANTZ is hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
// Written in 2010-2012 by Shane Saxon - contact@saxondigital.net
//
// Additional code contributed by: Mark Buchanan and Lucas Erickson.
//
// To the extent possible under law, the author(s) have dedicated all copyright 
// and related and neighboring rights to this software to the public domain
// worldwide. This software is distributed without any warranty.
//
// Released under the CC0 license, which is GPL compatible.
//
// You should have received a copy of the CC0 Public Domain Dedication along
// with this software (license file named COPYING.txt). If not, see
// http://creativecommons.org/publicdomain/zero/1.0/
//
//------------------------------------------------------------------------------

#include "npgldraw.h"

#include "npgeometry.h"

#include "../../npdata.h"
#include "../../npctrl.h"
#include "../npglut.h"


void DrawDefault (NPnodePtr node);
void DrawCamera (NPnodePtr node);
void DrawSurface (NPnodePtr node);
void DrawPoints (NPnodePtr node);
void DrawPin (int selected, NPnodePtr node, void* dataRef);
void DrawPinChild (NPnodePtr node, void* dataRef);
void DrawVideo (NPnodePtr node);
void DrawGrid (NPnodePtr node, void* dataRef);
void npDrawSort(void* dataRef);


//------------------------------------------------------------------------------
void npInitGLDraw(void* dataRef)
{
	pData data = (pData) dataRef;

//	InitTorus (dataRef);

	data->io.gl.pickPass = false;			//zz global init
}


//------------------------------------------------------------------------------
void DrawDefault (NPnodePtr node)
{
	return;
}


// draw an icon at the location of the camera in the scene
//------------------------------------------------------------------------------
void DrawCamera (NPnodePtr node)
{
	// draw icon of camera in the scene, but not the current viewport camera
	return;
}


//------------------------------------------------------------------------------
void DrawSurface (NPnodePtr node)
{
	// draw ground
	glPushMatrix();

	// position the node
	glTranslatef (node->translate.x, node->translate.y, node->translate.z);
	glRotatef (node->rotate.y, 0.0f, 0.0f, 1.0f);
	glRotatef (node->rotate.x, 1.0f, 0.0f, 0.0f);
	glRotatef (node->rotate.z, 0.0f, 0.0f, 1.0f);

	glScalef (node->scale.x, node->scale.y, node->scale.z);

	glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);

	glBegin(GL_QUADS);
		glVertex3f (-100.0f, -100.0f, 0.0f);
		glVertex3f (-100.0f,  100.0f, 0.0f);
		glVertex3f ( 100.0f,  100.0f, 0.0f);
		glVertex3f ( 100.0f, -100.0f, 0.0f);
	glEnd();

	glPopMatrix();
}


//------------------------------------------------------------------------------
void DrawPoints (NPnodePtr node)
{
	return;
}


//MB//

// converts the ID value to an RGB value for picking objects by color
//------------------------------------------------------------------------------
void npIDtoRGB( int id, int* r, int* g, int* b )
{
	int remain = id;
	
	*r = remain / (256*256);	remain = remain - *r * (256*256);
	*g = remain / 256;			remain = remain - *g * 256;
	*b = remain;
}

int npRGBtoID( int r, int g, int b )
{
	int id;
	id = (256*256)*r + 256*g + b;

	return id;
}


void PickPass( int picking, void* dataRef )
{
	pData data = (pData) dataRef;

	data->io.gl.pickPass = picking;
}

int npGetPickPass(void* dataRef)
{
	pData data = (pData) dataRef;

	return data->io.gl.pickPass;
}

//MB-End//

//------------------------------------------------------------------------------
void DrawPin (int selectedRootNode, NPnodePtr node, void* dataRef)
{
	int i = 0;
	int idRed = 0, idGrn = 0, idBlu = 0;

	pData data = (pData) dataRef;

	NPnodePtr nodeChild = data->map.activeNode;
	
	//if pickPass render scene with node ID encoded as color	//MB//
	if( data->io.gl.pickPass )
	{											//Convert nodeID to RGB triplet
		npIDtoRGB( node->id, &idRed, &idGrn, &idBlu );
		glColor4ub( idRed, idGrn, idBlu, 255 );	//ID of Node is encoded on the Pixel! In RGB Triplet   (Range: 0-  16.7 Million)
	}
	else
		glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a); //-


	glPushMatrix();

	// position the node
	glTranslatef (node->translate.x, node->translate.y, node->translate.z);

	// rotate
	glRotatef (node->rotate.y, 0.0f, 0.0f, -1.0f);
	glRotatef (node->rotate.x, -1.0f, 0.0f, 0.0f);
	glRotatef (node->rotate.z, 0.0f, 0.0f, -1.0f);

	//scale
	glScalef (node->scale.x, node->scale.y, node->scale.z);

	glLineWidth (node->lineWidth);

	//draw node
	if (node->geometry == kNPprimitiveSolidPin)
	{
		npGLSurface (true, node, data);	
		glTranslatef (0.0f, 0.0f, 5.0f);			//should this be here, debug, zz
	}
	else
	{
		glTranslatef (0.0f, 0.0f, 5.0f);
		npGLSurface (true, node, data);
	}

	glLineWidth(1.0f);	//zz

	if (node->selected || selectedRootNode)
	{
		glPushMatrix();

		if (selectedRootNode)
		{
			// slide down a little for scaling up to surround the pin
			glTranslatef (0.0f, 0.0f, -5.10f);
			glScalef (1.15f, 1.15f, 1.15f);

			if ( node->selected )
			{
				if( !data->io.gl.pickPass )
					glColor4f (1.0f, 1.0f, 0.0f, 0.8f);	// draw selected yellow
				npGLPrimitive (kNPprimitiveWirePin, 0.0f);
			}
			else if (node->branchLevel == 0)
			{
				if( !data->io.gl.pickPass )
					glColor4f ( 1.0f, 0.0f, 0.0f, 1.0f);	// current node red
				npGLPrimitive (kNPprimitiveWirePin, 0.0f);
			}

			if (node->branchLevel > 0 && !data->io.gl.pickPass)
			{
				if (nodeChild->selected)
					glColor4f (1.0f, 1.0f, 0.0f, 0.8f);	// draw selected yellow
				else
					glColor4f (1.0f, 0.0f, 0.0f, 1.0f);	// current node red
			}

			glTranslatef (0.0f, 0.0f, 5.0f);
			glScalef (1.0f, 1.0f, 0.66666667f);				// squish height, z 
			npGLPrimitive (kNPprimitiveWirePin, 0.0f);	// 2nd one will be above
		}
		else if (node->selected)
		{
			if( !data->io.gl.pickPass )
				glColor4f (1.0f, 1.0f, 0.0f, 0.8f);		// draw selected as yellow

			glTranslatef (0.0f, 0.0f, -5.10f);
			glScalef (1.05f, 1.05f, 1.05f);
			npGLPrimitive (kNPprimitiveWirePin, 0.0f);
		}

		glPopMatrix();
	}

	//should call DrawNode, that calls draw pin, move this to drawnodes...   debug, zz
	// recursively
	for (i=0; i < node->childCount; i++)
	{
		glPushMatrix();
			DrawPinChild (node->child[i], dataRef);
		glPopMatrix();
	}

	glPopMatrix();
}

// first draw child here, then break it up, draw nodes should iterate down, debug, zz
// the tree branches and take care of selected glows, etc...
// this routine should be generic regardless of branch level
//------------------------------------------------------------------------------
#define kNPinvertDeg 0.005555555555f		//equiv to 1/180

void DrawPinChild (NPnodePtr node, void* dataRef)
{
	int i = 0;
	int idRed = 0, idGrn = 0, idBlu = 0;

	pData data = (pData) dataRef;

	NPnodePtr parent = node->parent;
	NPpinPtr nodeData = (NPpinPtr) node->data;

	if (parent == NULL)								//take this out		debug zz	
		return;

	//Main Color for Rendering Pin
	if( data->io.gl.pickPass )
	{
		npIDtoRGB( node->id, &idRed, &idGrn, &idBlu );
		glColor4ub( idRed, idGrn, idBlu, 255 );	//ID of Node is encoded on the Pixel! In RGB Triplet   (Range: 0-  16.7 Million)
	}
	else
		glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);


	// position based on parent topo type
	if (parent->topo == kNPtopoCube)
	{
		//if zero then calculate using id/6 + 1		//update to use parent->childIndex, debug zz
		if (node->facet == 0)
			i = node->id - 6 * (node->id / 6) + 1;	// +1 to start at 1
		else
			i = node->facet;

		// rotate orthogonal to face based on facet then translate
		switch (i) // 
		{
			case 1 :
				glRotatef (90.0f, 0.0f, 1.0f, 0.0f);		//pos x	front
				glRotatef (90.0f, 0.0f, 0.0f, 1.0f);
				break;
			case 2 :
				glRotatef (90.0f, 0.0f, 1.0f, 0.0f);		//neg x	back
				glRotatef (180.0f, 1.0f, 0.0f, 0.0f);
				glRotatef (90.0f, 0.0f, 0.0f, 1.0f);
				break;
			case 3 :
				glRotatef (90.0f, 0.0f, 1.0f, 0.0f);		//pos y	right
				glRotatef (-90.0f, 1.0f, 0.0f, 0.0f);
				glRotatef (90.0f, 0.0f, 0.0f, 1.0f);
				break;
			case 4 :
				glRotatef (90.0f, 0.0f, 1.0f, 0.0f);		//neg y	left
				glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
				glRotatef (90.0f, 0.0f, 0.0f, 1.0f);
				break;
			case 5 : 									//pos z top, no rotation
			//	glRotatef (-90.0f, 0.0f, 0.0f, 1.0f);
				break;
			case 6 :
			//	glRotatef (-90.0f, 0.0f, 0.0f, 1.0f);
				glRotatef (180.0f, 0.0f, 1.0f, 0.0f);	//neg z	bottom
				break;
			default : break;
		}
		glTranslatef (0.0f, 0.0f, kNPcubeLength);	//translate from center to cube surface, explain, debug zz
		glScalef (kNPinvertDeg * kNPcubeLength, kNPinvertDeg * kNPcubeLength, 
					kNPinvertDeg * kNPcubeLength); //scale face +/- 180 deg
		
		//now position node specific coordinates
		glTranslatef (node->translate.x, node->translate.y, node->translate.z);

		glRotatef (node->rotate.y, 0.0f, 0.0f, -1.0f);
		glRotatef (node->rotate.x, -1.0f, 0.0f, 0.0f);
		glRotatef (node->rotate.z, 0.0f, 0.0f, -1.0f);
	//	glRotatef (node->rotate.z * kRADtoDEG, 0.0f, 0.0f, 1.0f);

		glScalef (90.0f / kNPcubeLength, 90.0f / kNPcubeLength, 90.0f / kNPcubeLength);	//scale node up to half parent size
		glScalef (node->scale.x, node->scale.y, node->scale.z);	//node scale
	}
	else if (parent->topo == kNPtopoSphere)
	{
	//	glRotatef (node->rotate.y * kRADtoDEG, 0.0f, 0.0f, 1.0f);	//CSV ver 1 debug zz
		//position
		glRotatef (node->translate.x + 180.0f, 0.0f, 0.0f, 1.0f);	//longitude, +180 so N = 0
		glRotatef (node->translate.y - 90.f, -1.0f, 0.0f, 0.0f);	//latitude

		//translate 1 unit to surface then convert node z local units
		glTranslatef ( 0.0f, 0.0f, 1.0f + node->translate.z / (4.0f*kRADtoDEG));
		
		//orientation
		glRotatef (node->rotate.y, 0.0f, 0.0f, -1.0f);
		glRotatef (node->rotate.x, -1.0f, 0.0f, 0.0f);
		glRotatef (node->rotate.z, 0.0f, 0.0f, -1.0f);

		//scale 1/2 size of parent
		glScalef (node->scale.x * 0.25f, node->scale.y * 0.25f,
					node->scale.z * 0.25f);	//node scale
//		glScalef (node->scale.x * 0.3536f, node->scale.y * 0.3536f,
//					node->scale.z * 0.3536f);	//node scale
	}
	else if (parent->topo == kNPtopoPin )
		//|| ( parent->topo == 0 && parent->geometry == kNPglutWireTorus) )	//default
	{
		//position the node vertically on the pin, note using translate.x
		if (node->branchLevel == 1)
		{
			glTranslatef (0.0f, 0.0f, node->translate.x / 37.22f);

			if ( node->topo == kNPtopoPin)// || (node->topo == 0 ))
				glScalef (node->scale.x * 0.5f, node->scale.y * 0.5f, node->scale.z * 0.5f);
			else
				glScalef (node->scale.x, node->scale.y, node->scale.z);
		}
		else
		{
			glTranslatef (0.0f, 0.0f, kNPpinHeight + node->translate.x / 37.22f);

			//scale by node size and additional factor to reduce size
			if ( node->topo == kNPtopoPin)// || (node->topo == 0 ))
				glScalef (node->scale.x * 0.5f, node->scale.y * 0.5f, node->scale.z * 0.5f);
			else
				glScalef (node->scale.x, node->scale.y, node->scale.z);
		}

		//offset from center after scaling, preserves ring offset when scaling
		glTranslatef (node->translate.z * .008333333f, //kRADtoDEG,	//scale 180 to ground
					 0.0f, 0.0f); // node->translate.y / -37.22f, 0.0f); //locked y, makes more sense when swtiching topos

//		if (parent->topo == kNPtopoPin)
//			glRotatef (90.0f, 0.0f, 0.0f, 1.0f);

		//rotate
	//	glRotatef (90.0f, 0.0f, 0.0f ,1.0f);	//orient x = 0 to north
		glRotatef (node->rotate.y, 0.0f, 0.0f, -1.0f);
		glRotatef (node->rotate.x, -1.0f, 0.0f, 0.0f);
		glRotatef (node->rotate.z, 0.0f, 0.0f, -1.0f);
	}
	else if (node->branchLevel > 1 || parent->topo == kNPtopoTorus) //topo torus
	{
		//scale up 1.5X to match torus geometry size inner radius
		glScalef (kNPtorusRadius, kNPtorusRadius, kNPtorusRadius); 
		
		//position at torus outer radius, inside center of tube
		glRotatef (node->translate.x + 90.0f, 0.0f, 0.0f, 1.0f);
//		glRotatef (node->translate.x, 0.0f, 0.0f, 1.0f);		//longitude
		glTranslatef (1.0f, 0.0f, 0.0f);	//translate to center of tube

		//?
//		glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
		
		//treat null as torus, later make this depend on the geometry
		if (node->topo == kNPtopoTorus || node->topo == 0)
	//		|| node->geometry == kNPglutWireTorus
	//		|| node->geometry == kNPglutSolidTorus )
		{
			glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
			glRotatef (-90.0f, 0.0f, 0.0f, 1.0f);
			glRotatef (node->translate.y, 0.0f, 0.0f, 1.0f);	//latitude
			//translate.z based on scale.x and is converted from deg 57.29...
			glTranslatef (0.0f, 0.1591f * node->scale.x * node->translate.z 
						  / kRADtoDEG, 0.0f);
		}
		else
		{	
			glRotatef (90.0f, 0.0f, 1.0f, 0.0f);		
			glRotatef (node->translate.y, 0.0f, -1.0f, 0.0f);	//latitude

			glRotatef (90.0f, 0.0f, 0.0f, 1.0f);				//orient North
			//translate 1 unit to surface then convert node z local units
			//uses parent->ratio for torus inner radius and 1.5f for scaling factor
			if (node->topo == kNPtopoPin)// || node->geometry == kNPprimitiveSolidPin) // kNPtopoTorus zz
			{
				glTranslatef ( 0.0f, 0.0f,
					parent->ratio + (node->translate.z / kRADtoDEG) );
	//			glRotatef (node->translate.y, 0.0f, -1.0f, 0.0f);	//latitude
				glScalef (0.5f, 0.5f, 0.5f);	//results in 1/4 parent
			}
			else //if (node->topo == kNPtopoNull)
			{
				glTranslatef (0.0f, 0.0f, node->translate.z / kRADtoDEG);
//				glRotatef (node->translate.y, 0.0f, -1.0f, 0.0f);	//latitude
			}

		}
		//orientation
		glRotatef (node->rotate.y, 0.0f, 0.0f, -1.0f);
		glRotatef (node->rotate.x, -1.0f, 0.0f, 0.0f);
		glRotatef (node->rotate.z, 0.0f, 0.0f, -1.0f);

		//scale 1/2 size of parent, smaller is good for torus .5 / 1.5
		glScalef (node->scale.x * 0.33333333f, node->scale.y * 0.33333333f,
					node->scale.z * 0.33333333f);	//node scale

/*
		//position
		glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
		glRotatef (node->rotateVec.angle, 0.0f, 1.0f, 0.0f);		//x	debug zz
			
		glTranslatef ( (node->translate.x * node->scale.x) + 1.5f,
						node->translate.y * node->scale.y,
						0.0f );								// z not translated

		//orientation
		glRotatef (node->rotate.y * kRADtoDEG, 0.0f, 0.0f, 1.0f);		//y, debug zz
		glRotatef (node->rotate.z * kRADtoDEG, 1.0f, 0.0f, 0.0f);		//y, debug zz

		glScalef ( 0.5f * node->scale.x, 0.5f * node->scale.y, 0.5f * node->scale.z);
*/	
	}
//	else if(node->geometry == kNPprimitiveSolidPin)// process pins, DrawPin(node, dataRef);
//	{
//		DrawPin (false, node, dataRef);
//	}	
	else //assume primary torus with parent as pin
	{
		if (node->branchLevel == 1)
		{
			if (parent->topo == 0 || parent->topo == kNPtopoPin)
			{	
				glTranslatef (0.0f, 0.0f, node->translate.x / 37.22f);
				if (node->topo == kNPtopoPin)
					glScalef (0.5f, 0.5f, 0.5f);
			}
			else if (parent->topo == kNPtopoTorus)
			{
				glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
				//translate.z based on scale.x and is converted from deg 57.29...
				glTranslatef (0.3722f * node->scale.x * node->translate.z 
							  / kRADtoDEG, 0.0f, 0.0f);
				glRotatef (node->translate.y, 0.0f, 0.0f, 1.0f);	//latitude
			}
			else
				glTranslatef (node->translate.x * node->scale.x,
						node->translate.y * node->scale.y,
						node->translate.z );				// z is not scaled
		}
		else	
			glTranslatef (node->translate.x * node->scale.x,
						node->translate.y * node->scale.y,
						node->translate.z );				// z is not scaled
		
	//	glRotatef (node->rotate.y * kRADtoDEG, 0.0f, 0.0f, 1.0f);		//y, debug zz
	//	glRotatef (90.0f, 0.0f, 0.0f, 1.0f);	//90 deg orient North
	
		//orientation
		glRotatef (node->rotate.y, 0.0f, 0.0f, -1.0f);
		glRotatef (node->rotate.x, -1.0f, 0.0f, 0.0f);
		glRotatef (node->rotate.z, 0.0f, 0.0f, -1.0f);
		
		glScalef (node->scale.x, node->scale.y, node->scale.z);	//root child
	}


	// draw the node unless hidden
	if (!node->hide)
	{
		glLineWidth(node->lineWidth);

		if (node->branchLevel > 1)
		{
			if (node->topo == kNPtopoPin && node->geometry != kNPprimitiveSolidPin)
			{
				glTranslatef (0.0f, 0.0f, 5.0f);
				npGLSurface (true, node, data);
				glTranslatef (0.0f, 0.0f, -5.0f);
			}
			else
			{
				npGLSurface (true, node, data);
			//	glTranslatef (0.0f, 0.0f, 5.0f);
			}
		}
		else
			npGLSurface (true, node, data);	

		glLineWidth(1.0f);
	}

	// draw a selection status wireframes slightly scaled up on top...
	if (node->selected)
	{	
		glPushMatrix();
			// draw nodeID color on pickPass to allow picking on the wireframe
			if( !data->io.gl.pickPass )
				glColor4f (0.8f, 0.8f, 0.0f, 1.0f);			//draw yellow
			glScalef (1.03f, 1.03f, 1.03f);
			npGLPrimitive (kNPglutWireTorus, node->ratio);
		glPopMatrix();
	}

	// draw a wireframe slightly scaled up on top...
	if (node == data->map.selectedPinNode)
	{	
		glPushMatrix();
			// draw nodeID color on pickPass to allow picking on the wireframe
			if( !data->io.gl.pickPass )
			{
				if (node->freeze)
					glColor4f (0.0f, 1.0f, 1.0f, 0.6f);			//draw cyan
				else
					glColor4f (1.0f, 0.0f, 0.0f, 1.0f);		//draw red
			}
			glScalef (1.04f, 1.04f, 1.04f);
			npGLPrimitive (kNPglutWireTorus, node->ratio);
		glPopMatrix();
	}
	
	// draw the child nodes recursively
	for (i=0; i < node->childCount; i++)
	{
		glPushMatrix();
			DrawPinChild (node->child[i], data);
		glPopMatrix();	
	}
}


//------------------------------------------------------------------------------
void DrawVideo (NPnodePtr node)
{
	return;
}


//------------------------------------------------------------------------------
void DrawGrid (NPnodePtr node, void* dataRef)
{
	int			i = 0;
	int			j = 0;
	int			k = 0;
	int			idRed = 0, idGrn = 0, idBlu = 0;
	GLfloat		x = 0.0f;
	GLfloat		y = 0.0f;
	GLfloat		z = 0.0f;
	NPfloatXYZ	length;

	pData data = (pData) dataRef;
	NPgridPtr grid = (NPgridPtr) node->data;
	NPnodePtr camNode = data->map.activeCam;

	//MB-TEXTURE
	static GLfloat sgenparams[] = { +1.0,  0.0,  0.0,  0.0 };
	static GLfloat tgenparams[] = {  0.0, +1.0,  0.0,  0.0 };
	//MB-END

	//if pickPass and pickMode not grid then don't draw grid
	if (data->io.gl.pickPass)// && data->map.nodeRootIndex != kNPnodeRootGrid) debug, zz
		return;

	//turn two-sided rendering of polygon faces off
	if (node->branchLevel > 0)
		glDisable (GL_CULL_FACE);

	glDisable (GL_LIGHTING);	//draw 100% ambient white
	
	glPushMatrix();				//restore after calling child nodes

	//MB-TEXTURE
	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL ); //GL_MODULATE
	glTexGeni( GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR ); //GL_SPHERE_MAP
	glTexGeni( GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR );
	glTexGenfv( GL_S, GL_OBJECT_PLANE, sgenparams );
	glTexGenfv( GL_T, GL_OBJECT_PLANE, tgenparams );
	glDisable (GL_TEXTURE_GEN_S);
	glDisable (GL_TEXTURE_GEN_T);
	//MB-END

	//used to calculate grid center based on segment count and spacing
	length.x = grid->spacing.x * node->segments.x;
	length.y = grid->spacing.y * node->segments.y;

	// position node
	glTranslatef (node->translate.x, node->translate.y, node->translate.z);

	glRotatef (node->rotate.y, 0.0f, 0.0f, -1.0f);
	glRotatef (node->rotate.x, -1.0f, 0.0f, 0.0f);
	glRotatef (node->rotate.z, 0.0f, 0.0f, -1.0f);
		
	glScalef (node->scale.x, node->scale.y, node->scale.z);

	//center based on length
	glTranslatef (length.x * -0.5f, length.y * -0.5f, 0.0f);

	// if pickPass then set color by id
	if (data->io.gl.pickPass)
	{
		npIDtoRGB (node->id, &idRed, &idGrn, &idBlu);

		printf("r: %d   g: %d   b: %d\n", idRed, idGrn, idBlu);
		glColor4b (idRed, idGrn, idBlu, 255);		//draw quad color mapped as id
	}

	//Draw Texture Quad
	if (node->textureID)	//allow for click on background tex's for mouse flying
	{
		//if not pickPass then enable and bind texture, set color to white
		if (!data->io.gl.pickPass)
		{
			glEnable (GL_TEXTURE_2D);
			glBindTexture (GL_TEXTURE_2D, node->textureID);
			if (node->branchLevel > 0)	//child grids use opcacity for textures
				glColor4ub (255, 255, 255, node->color.a);
			else
				glColor4f (1.0f, 1.0f, 1.0f, 1.0f);	 //root grid is fully opaque
		}

		//draw the quad
		glBegin (GL_QUADS);
			glTexCoord2d(0.0,0.0); glVertex3d(0.0f,  0.0f, 0.0f);	 // center quad
			glTexCoord2d(1.0,0.0); glVertex3d(length.x, 0.0f, 0.0f);
			glTexCoord2d(1.0,1.0); glVertex3d(length.x, length.y, 0.0f);
			glTexCoord2d(0.0,1.0); glVertex3d(0.0f,  length.y, 0.0f);
		glEnd();

		glDisable ( GL_TEXTURE_2D );
	}

	//draw grid lines unless hidden
	if (!node->hide)
	{
		// if not pickPass then set the node color
		if (!data->io.gl.pickPass)
			glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);

		//draw grid lines on top of texture map
		glDisable (GL_DEPTH_TEST);
		glLineWidth (node->lineWidth);

		//draw grid lines
		glBegin (GL_LINES);
			for (i=0; i <= node->segments.z; i++)
			{
				for (j=0; j <= node->segments.x; j++)
				{
					glVertex3f (x, 0.0f, z);
					glVertex3f (x, length.y, z);
					x += grid->spacing.x;
				}
				for (j=0; j <= node->segments.y; j++)
				{
					glVertex3f (0.0f, y, z);
					glVertex3f (length.x, y, z);
					y += grid->spacing.y;
				}

				//if activeNode then draw yellow selection box
				if (node == data->map.activeNode && !data->io.gl.pickPass)
				{
					glColor4ub(255,255,0,192);	//yellow, 75% opacity

					if (node == data->map.activeNode)
					{
						glVertex3f (0.0f, 0.0f, z);			//side one
						glVertex3f (length.x, 0.0f, z);

						glVertex3f (length.x, 0.0f, z);		//two
						glVertex3f (length.x, length.y, z);

						glVertex3f (length.x, length.y, z);	//three
						glVertex3f (0.0f, length.y, z);

						glVertex3f (0.0f, length.y, z);		//four
						glVertex3f (0.0f, 0.0f, z);

						glVertex3f (length.x * 0.5f, 0.0f, z);	 //cross hairs
						glVertex3f (length.x * 0.5f, length.y, z);

						glVertex3f (0.0f, length.y * 0.5f, z);	 //cross hairs
						glVertex3f (length.x, length.y * 0.5f, z);
					}
						// restore node color 
					glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);
				}
				x = 0.0f;						//reset x and y
				y = 0.0f;
				z += grid->spacing.z;			//set the z for 3D stacked grids
			}
		glEnd();

		glLineWidth (1.0f);
		glEnable (GL_DEPTH_TEST);
	}

	//turn two-sided rendering of polygon faces off
	if (node->branchLevel > 0)
		glEnable (GL_CULL_FACE);

	// if not pickPass then restore lighting
	if (!data->io.gl.pickPass)
		glEnable (GL_LIGHTING);

	//undo center based on length
	glTranslatef (length.x * 0.5f, length.y * 0.5f, 0.0f);

	//do not want scaling of root grid to effect children, single case
	if (node == data->map.node[kNPnodeRootGrid])
		glPopMatrix();				//restore for root grid only

	//recursively traverse and draw children
	for (i=0; i < node->childCount; i++)
		DrawGrid(node->child[i], data);

	//for all other nodes the transformations are inherited
	if (node != data->map.node[kNPnodeRootGrid])
		glPopMatrix();				// restore after traversing child nodes
}


// draw 2D overlays including grids, text, messages, highlight objects, etc...
//------------------------------------------------------------------------------
void npDrawHUD (void* dataRef)
{
	return;
}


// sets the node RGB color based on index, preserves existing alpha value
//------------------------------------------------------------------------------
void SetIndexColor (NPubyteRGBA *color, int *colorIndex)
{	
	int index = *colorIndex;
									//kNPpaletteSize stored in nptypes.h
	static const GLubyte colorPallete[kNPpaletteSize][3] = {
											{50,  101, 101},
											{0,   255, 0},
											{255,   0, 0},
											{0,     0, 255},
											{255, 255, 0},
											{152,   0, 255},
											{255, 168, 0},
											{0,   255, 255},
											{255,   0, 255},
											{0,   153, 0},
											{185, 153, 102},
											{255, 180, 255},
											{0,   152, 255},
											{185, 255, 0},
											{152,   0, 0},
											{127, 127, 127},
											{127, 127, 255},
											{197,  82, 0},
											{0,     0, 0},
											{255, 255, 255}
											};
	//re-maps out of bounds colorIndex
	if (index >= kNPpaletteSize)
	{
		while (index >= kNPpaletteSize)
			index -= kNPpaletteSize;	//subtracts pallete sized increments
		*colorIndex = index;
	}
	else if (index < 0)
	{
		while (index < 0)
			index += kNPpaletteSize;
		*colorIndex = index;			//adds pallete sized increments		
	}

	//now set the color
	color->r = colorPallete[index][0];
	color->g = colorPallete[index][1];
	color->b = colorPallete[index][2];
}




//MB-Transp
//------------------------------------------------------------------------------
//Distance between two points
float npDistance( NPfloatXYZ vecA, NPfloatXYZ vecB )
{
	float x = vecA.x - vecB.x;
	float y = vecA.y - vecB.y;
	float z = vecA.z - vecB.z;
	
	float len = x*x + y*y + z*z;
	len = (float)sqrt( len );
	return len;
}

//------------------------------------------------------------------------------
//QSort Compare function
int comparePinsDistanceFromCamera (const void* a, const void* b)
{
	NPnodePtr* pinAptr = (NPnodePtr*)a;
	NPnodePtr* pinBptr = (NPnodePtr*)b;
	NPnodePtr pinA = *pinAptr;
	NPnodePtr pinB = *pinBptr;
	if( pinA == NULL && pinB == NULL )	//Null element in pinA & pinB: Nothing or Grid or Camera
		return 0;						//These are "equal". Both should be at the beginning of the list.
	if( pinA == NULL )					//Null element in pinA: Nothing or Grid or Camera
		return -1;						//pinA is at beginning of list... therefore pinA is    "less than" pinB. Return -1
	if( pinB == NULL )					//Null element in pinB: Nothing or Grid or Camera
		return +1;						//pinB is at beginning of list... therefore pinA is "greater than" pinB. Return +1
	
	if( pinA->distFromCamera   >  pinB->distFromCamera )	//pinA is furter from camera than pinB.
		return -1;											//Therefore should be at the beginning of the list. ie "less than"... Return -1
	if( pinA->distFromCamera   <  pinB->distFromCamera )	//pinA is closer from camera than pinB.
		return +1;											//Therefore should be at the end of the list.    ie "greater than"... Return +1
	if( pinA->distFromCamera  ==  pinB->distFromCamera )	//pinA & pinB are exactly the same distance from the camera. (impossible-ish)
		return 0;											//Return 0;

	//return ( *(int*)a - *(int*)b );
	return -1;
}


//------------------------------------------------------------------------------
void npDrawSort(void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	NPnodePtr node = NULL;
	NPnodePtr camNode = data->map.activeCam;

	int nodeRootCount = data->map.nodeRootCount;

	NPfloatXYZ camVec = camNode->translate;

	//update root node distance from camera
	for( i = 0; i < nodeRootCount; i++ )
	{
		if( i < kNPnodeRootPin )	//Nothing, Camera, Grid
		{
			data->map.sort[i] = NULL;
			continue;
		}
		
		node = data->map.node[i];
		node->distFromCamera = npDistance( camVec, node->translate );

		data->map.sort[i] = node;
	}

	//create z-sorted list, back to front
	//QuickSort based on Distance From Camera, back to front z-sorted
	qsort( data->map.sort, nodeRootCount, sizeof(NPnodePtr),
			comparePinsDistanceFromCamera );

}

//------------------------------------------------------------------------------
void npDrawNodes (void* dataRef)
{
	int i = 0;
	int selected = false;

	NPnodePtr node = NULL;
	pData data = (pData) dataRef;

	int count = data->map.nodeRootCount;

	//sort nodes, back to front z-sort for transparency
	npDrawSort(data);

	//draw all nodes
	for( i=0; i < count; i++ )
	{
		if( i < kNPnodeRootPin )
			continue; // node = data->map.node[i];	//first draw camera and grids
		//else
		
		node = data->map.sort[i];	//draw in sorted order
		
		if (node == data->map.node[data->map.nodeRootIndex]) //use seletedIndex? debug zz
			selected = true;
		else
			selected = false;



		if (!node->hide)				//if not hidden draw node
			switch (node->type)
		{
			case kNodeDefault	: DrawDefault(node); break;
			case kNodeCamera	: DrawCamera (node); break;
			case kNodeSurface	: DrawSurface(node); break;
			case kNodePoints	: DrawPoints (node); break;
			case kNodePin		: DrawPin (selected, node, data); break;
			case kNodeVideo		: DrawVideo  (node); break;
			case kNodeGrid		: DrawGrid   (node, data); break;
		}
	}
	//draw grids last for transparency to work right
	DrawGrid (data->map.node[kNPnodeRootGrid], data);
}

	

void npDrawNode (NPnodePtr node, void* data)
{
	switch (node->type)
	{
		case kNodeDefault	: DrawDefault	(node); break;
		case kNodeCamera	: DrawCamera	(node); break;
		case kNodeSurface	: DrawSurface	(node); break;
		case kNodePoints	: DrawPoints	(node); break;
		case kNodePin		: DrawPin		(0, node, data); break;
		case kNodeVideo		: DrawVideo		(node); break;
		case kNodeGrid		: DrawGrid		(node, data); break;
	}
}

//MB-LABEL
void DrawSimpleTorus(void)
{
	npGlutPrimitive (kNPglutSolidTorus);
}

